package org.thibault.android.clock.binary;

import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;

import android.app.PendingIntent;
import android.app.Service;
import android.appwidget.AppWidgetManager;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PixelXorXfermode;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region;
import android.graphics.Typeface;
import android.graphics.drawable.Drawable;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.preference.PreferenceManager;
import android.util.Log;
import android.util.SparseArray;
import android.util.SparseIntArray;
import android.widget.RemoteViews;
import android.widget.Toast;

public class Updater extends Service
{
	private static final String WIDGET_ID_COUNT_KEY="WIDGET_ID_COUNT_KEY";
	private static final String WIDGET_ID_KEY="WIDGET_ID_KEY";
	private static final String WIDGET_RADIUS_KEY="WIDGET_RADIUS_KEY";
	private static final String WIDGET_WITH_TEXT_KEY="WIDGET_WITH_TEXT_KEY";
	private static final String WIDGET_STYLE_ID_KEY="WIDGET_STYLE_ID_KEY";
	
	private boolean paused = false;
	
	static void say(Context context, String msg){
		say(msg);
		//Toast t=Toast.makeText(context,msg,Toast.LENGTH_SHORT);
		//t.show();
	}

	static void say(String msg){

		//Log.d("BinaryClock.Updater",msg);
	}

	private BroadcastReceiver receiver = null;

    public void onCreate(){
    	say(getApplicationContext(), "onCreate");
        // REGISTER RECEIVER THAT HANDLES SCREEN ON AND SCREEN OFF LOGIC
        IntentFilter filter = new IntentFilter(Intent.ACTION_SCREEN_ON);
        filter.addAction(Intent.ACTION_SCREEN_OFF);
        filter.addAction(Intent.ACTION_TIME_CHANGED);
        filter.addAction(Intent.ACTION_DATE_CHANGED);
        filter.addAction(Intent.ACTION_TIMEZONE_CHANGED);
        filter.addAction(Intent.ACTION_TIME_TICK);
        receiver = new ScreenReceiver();
        registerReceiver(receiver, filter);
    }
    
    public void onDestroy() {
		unregisterReceiver(receiver);
    	receiver = null;
    }

	private class ScreenReceiver extends BroadcastReceiver {
	    @Override
	    public void onReceive(Context context, Intent intent) {
	        if (intent.getAction().equals(Intent.ACTION_SCREEN_OFF)) {
	            paused = true;
	        }else{
	        	if (intent.getAction().equals(Intent.ACTION_SCREEN_ON)) {
	        		paused = false;
	        	}
	        }
        	if (!paused)
        		doUpdate();
	    }
	}

	public static class WidgetProperties {
		public final int diameter;
		public final boolean withText;
		public final int styleId;
		public WidgetProperties(int _diameter, boolean _withText, int _styleId) {
			diameter=_diameter;
			withText=_withText;
			styleId=_styleId;
		}
		
		public void save(SharedPreferences.Editor edit, int id) {
			edit.putInt(WIDGET_RADIUS_KEY+"_"+id, diameter);
			edit.putBoolean(WIDGET_WITH_TEXT_KEY+"_"+id, withText);
			edit.putInt(WIDGET_STYLE_ID_KEY+"_"+id, styleId);
		}
		
		public static WidgetProperties load(SharedPreferences p, int id) {
			final int _diameter=p.getInt(WIDGET_RADIUS_KEY+"_"+id, -1);
			if (_diameter<0) return null;
			final boolean _withText=p.getBoolean(WIDGET_WITH_TEXT_KEY+"_"+id, false);
			final int _styleId=p.getInt(WIDGET_STYLE_ID_KEY+"_"+id, StyleRegistry.DEFAULT);
			return new WidgetProperties(_diameter,_withText,_styleId);
		}
	}
	
	public static final String EXTRA_WIDGET_IDS="EXTRA_WIDGET_IDS";
    private Set<Integer> appWidgetIds=new HashSet<Integer>();
	private int lastTick=-1;
	private final SparseArray<WidgetProperties> properties=new SparseArray<WidgetProperties>();
	public static float density = -1;
	private static Drawable greenStone = null, blueStone = null;
	
	private void saveWidgets() {
		SharedPreferences p=PreferenceManager.getDefaultSharedPreferences(this);
		SharedPreferences.Editor edit=p.edit();
		edit.putInt(WIDGET_ID_COUNT_KEY, properties.size());
		int i=0;
		for (i=0; i<properties.size(); i++) {
			final int id=properties.keyAt(i);
			edit.putInt(WIDGET_ID_KEY+"_"+i, id);
			properties.get(id).save(edit, id);
			i++;
		}
		edit.commit();
	}
	
	private void loadWidgets() {
		SharedPreferences p=PreferenceManager.getDefaultSharedPreferences(this);
		final int N=p.getInt(WIDGET_ID_COUNT_KEY, 0);
		appWidgetIds.clear();
		properties.clear();
		for (int i=0; i<N; i++) {
			final int id=p.getInt(WIDGET_ID_KEY+"_"+i, -1);
			if (id<0) continue;
			appWidgetIds.add(id);
			WidgetProperties prop=WidgetProperties.load(p, id);
			if (prop!=null)
				properties.put(id, prop);
		}
	}
	
	private void addWidget(int id) {appWidgetIds.add(id); lastTick=-1;}
	private void delWidget(int id) {appWidgetIds.remove(id); properties.delete(id);}
	private void delAllWidgets() {
		appWidgetIds.clear();
		properties.clear();
	}
	private void setWidgetProperties(int id, int width, int height,
									 boolean withText, int styleId) {
		int diameter=min(width,height);
		properties.put(id, new WidgetProperties(diameter,withText,styleId));
	}

	private void setWidgetProperties(int id, boolean withText, int styleId) {
		setWidgetProperties(id, 0, 0, withText,styleId);
	}
	
	public boolean isEmpty() {return appWidgetIds.isEmpty();}
	
	public IBinder onBind(Intent p1) {return null;}
	
    public synchronized void onStart(Intent intent, int startId) 
    {
    	say(getApplicationContext(),"onStart(): "+((intent==null) ? "null" : intent.getAction()));
    	
    	if ((density<0) && (intent!=null))
    		density=intent.getFloatExtra(Widget.CLOCK_SCREEN_DENSITY, -1);
    	
        if ((intent==null) || (intent.getAction().equalsIgnoreCase(Widget.CLOCK_START))) {
        	if (intent!=null) {
        		int[] ids=intent.getIntArrayExtra(Widget.CLOCK_WIDGET_IDS);
        		for (int id: ids) {
        			say("Widget in start intent: "+id);
        			SharedPreferences p=PreferenceManager.getDefaultSharedPreferences(this);
        			WidgetProperties props=WidgetProperties.load(p, id);
        			boolean withText=intent.getBooleanExtra(Widget.CLOCK_WIDGET_WITH_TEXT,
        													(props==null) ? false : props.withText);
        			int styleId=intent.getIntExtra(Widget.CLOCK_WIDGET_STYLE_ID,
        										     ((props==null)
        										      ? StyleRegistry.DEFAULT
        										      : props.styleId));
        			Bundle widgetOptions=intent.getBundleExtra(Widget.CLOCK_WIDGET_OPTIONS+"_"+id);
        			addWidget(id);
        			if (widgetOptions!=null) {
        				int width=widgetOptions.getInt(AppWidgetManager.OPTION_APPWIDGET_MAX_WIDTH);
        				int height=widgetOptions.getInt(AppWidgetManager.OPTION_APPWIDGET_MAX_HEIGHT);
        				if ((width>0) && (height>0)) {
            				say("Widget initial size: "+width+"x"+height);
        					setWidgetProperties(id,width,height,withText,styleId);
        				} else {
        					say("Widget is 0x0; does not actually exist");
        					setWidgetProperties(id,-1,-1,withText,styleId);
        				}
        			} else {
        				setWidgetProperties(id,200,200,withText,styleId);
        			}
        		}
        		saveWidgets();
        	} else {
        		say("Restoring widget data");
        		loadWidgets();
        	}
        	doUpdate();
        	paused = false;
        } else {
        	if(intent.getAction().equalsIgnoreCase(Widget.CLOCK_STOP)) {
        		paused = true;
        		delAllWidgets();
            	saveWidgets();
        		stopSelf();
        	} else {
        		if (intent.getAction().equalsIgnoreCase(Widget.CLOCK_RESIZE)) {
        			final int id=intent.getIntExtra(Widget.CLOCK_WIDGET_ID, -1);
        			final WidgetProperties props=properties.get(id);
        			final boolean withText=
        					((props!=null)
        					 ? props.withText
        					 : intent.getBooleanExtra(Widget.CLOCK_WIDGET_WITH_TEXT, false)
        					 );
        			final int styleId=
        					((props!=null)
        					 ? props.styleId
        					 : intent.getIntExtra(Widget.CLOCK_WIDGET_STYLE_ID, StyleRegistry.DEFAULT)
        					 );
        			Bundle widgetOptions=intent.getBundleExtra(Widget.CLOCK_WIDGET_OPTIONS);
        			int width=widgetOptions.getInt(AppWidgetManager.OPTION_APPWIDGET_MAX_WIDTH);
        			int height=widgetOptions.getInt(AppWidgetManager.OPTION_APPWIDGET_MAX_HEIGHT);
        			addWidget(id);
        			setWidgetProperties(id,width,height,withText,styleId);
                	saveWidgets();
            		lastTick=-1;
            		doUpdate();
        		} else {
        			if (intent.getAction().equalsIgnoreCase(Widget.CLOCK_DELETE)) {
        				int[] ids=intent.getIntArrayExtra(Widget.CLOCK_WIDGET_IDS);
        				for (int id: ids)
        					delWidget(id);
        				saveWidgets();
        			}
        		}
        	}
        }
    }
    
    private static StyleRegistry styleRegistry=new StyleRegistry();
    
	public static Bitmap mkBitmap(final int tick, final int _diameter,
								  boolean withText, int styleId,
								  Resources resources) {
		if (_diameter<0) return null;
		final int diameter=(_diameter==0 ? 200 : _diameter);
		//say("mkBitmap("+tick+","+diameter+")");
		Style style=styleRegistry.getStyle(styleId, resources);
		Bitmap b=Bitmap.createBitmap(diameter,diameter,Bitmap.Config.ARGB_8888);
		Canvas c=new Canvas(b);
		style.drawClock(tick, diameter, withText, c);
		return b;
	}
	
	public static int min(int a, int b) {
		if (a<b) return a; else return b;
	}
	
	public static int getTick() {
		Calendar now=Calendar.getInstance();
		return now.get(Calendar.HOUR)*60+now.get(Calendar.MINUTE);
	}
	
	private synchronized void doUpdate() {
	try {		
			int tick=getTick();
			//say(this,"lastTick=="+lastTick+"; tick=="+tick);
			if (tick==lastTick) return;
			
			for (int id: appWidgetIds) {
			
				WidgetProperties props=properties.get(id);
				if (props==null) {
					say("Skipping widget with no props "+id);
					continue;
				}
					
				final int d=props.diameter;
				if (d<=0) {
					say("Skipping 0x0 widget "+id);
					continue;
				}
				//say("Updating widget "+id);
				final int key=(d<<1) | (props.withText ? 1 : 0);
			
				Bitmap b=mkBitmap(tick,d,props.withText,props.styleId,getResources());

				RemoteViews views = new RemoteViews(getPackageName(),
													R.layout.widget);
				views.setImageViewBitmap(R.id.binaryClock,b);
				
				Intent intent = new Intent(getApplicationContext(), Configure.class);
				intent.putExtra(AppWidgetManager.EXTRA_APPWIDGET_ID, id);
				intent.putExtra(Configure.RECONFIGURING, true);
				/*{
					boolean reconfiguring=intent.getBooleanExtra(Configure.RECONFIGURING, false);
					say("reconfiguring=="+reconfiguring);
				}*/
				PendingIntent pending = PendingIntent.getActivity(getApplicationContext(), id, intent, 0);
				views.setOnClickPendingIntent(R.id.binaryClock, pending);

				AppWidgetManager.getInstance(this).updateAppWidget(id, views);
			}
			lastTick=tick;
	} catch (RuntimeException re) {
		say(getApplicationContext(),"Exception in doUpdate(): "+re);
		throw re;
	}
	}
}
